import { CONSTRUCTOR_FUNCTION_NAME, formatType, forwardArguments, INDENT, INTERNAL, SELF_ARG } from './common';
import { Struct } from './struct';
import { structuresWithVTableAddress } from './vtable';

// Allow Making Custom Structures
export class CustomWrapperGenerator {
    readonly #struct: Struct;

    // Constructor
    constructor(struct: Struct) {
        this.#struct = struct;
    }

    // Getter
    #getBase() {
        return `${INTERNAL}CustomBase<${this.#struct.name}, ${this.#struct.getVTable().getName()}>`;
    }
    #getMethods() {
        const out = [];
        out.push(this.#struct.getMainConstructor());
        const methods = this.#struct.getVTable().getMethods(false);
        for (const method of methods) {
            if (method) {
                out.push(method);
            }
        }
        return out;
    }
    #getName() {
        return 'Custom' + this.#struct.name;
    }

    // Generate Header
    generate() {
        let out = '';

        // Check
        if (!structuresWithVTableAddress.includes(this.#struct.name)) {
            throw new Error('Generating A Custom Wrapper Requires The VTable Address To Be Specified');
        }

        // Structure
        const base = this.#getBase();
        out += `struct ${this.#getName()} : ${base} {\n`;

        // Methods
        const methods = this.#getMethods();
        for (const method of methods) {
            // Generate Method
            const name = method.shortName;
            out += INDENT;
            if (name === CONSTRUCTOR_FUNCTION_NAME) {
                out += 'explicit ' + this.#getName();
            } else {
                out += 'virtual ' + formatType(method.returnType) + name;
            }
            out += method.getArgs(false) + ';\n';
        }

        // VTable
        out += 'private:\n';
        out += INDENT + `virtual void ${INTERNAL}patch_vtable(${base}::${INTERNAL}VTable *) override;\n`;
        out += '};\n';

        // Return
        return out;
    }

    // Generate Code
    generateCode() {
        let out = '';

        // Methods
        const methods = this.#getMethods();
        for (const method of methods) {
            const name = method.shortName;
            const isConstructor = name === CONSTRUCTOR_FUNCTION_NAME;
            // Generate Method
            if (!isConstructor) {
                out += formatType(method.returnType);
            }
            out += this.#getName() + '::';
            out += isConstructor ? this.#getName() : name;
            const args = method.getArgs(false);
            out += args + ' {\n';
            out += INDENT;
            const forwardedArgs = forwardArguments(args, ['this->self']);
            if (!isConstructor) {
                // Virtual Method
                if (method.doesReturnValue()) {
                    out += 'return ';
                }
                out += `${INTERNAL}VTable::base->${name}${forwardedArgs}`;
            } else {
                // Constructor
                out += method.getDirectCall() + forwardedArgs;
            }
            out += ';\n';
            if (isConstructor) {
                out += `${INDENT}${INTERNAL}set_vtable();\n`;
            }
            out += '}\n';
        }

        // VTable
        const base = this.#getBase();
        out += `void ${this.#getName()}::${INTERNAL}patch_vtable(${base}::${INTERNAL}VTable *vtable) {\n`;
        out += `${INDENT}${base}::${INTERNAL}patch_vtable(vtable);\n`;
        for (const method of methods) {
            const name = method.shortName;
            if (name === CONSTRUCTOR_FUNCTION_NAME) {
                continue;
            }
            // Patch Entry
            out += `${INDENT}vtable->${name} = []${method.getArgs()} {\n`;
            out += INDENT + INDENT;
            if (method.doesReturnValue()) {
                out += 'return ';
            }
            out += `custom_get<${this.#getName()}>(${SELF_ARG})->${name}${forwardArguments(method.getArgs(false), [])};\n`;
            out += INDENT + '};\n';
        }
        out += '}\n';

        // Return
        return out;
    }
};